from crmify.settings import crmify_settings
import os


def get_backend():
    backend_cls = crmify_settings.BACKEND
    return backend_cls(**crmify_settings.BACKEND_AUTH)


crm_backend = get_backend()


from crmify.backends.base import CRMBackend
from crmify.mixins import FieldMapperMixin
from crmify.lib.insightly import Insightly as InsightlyAPI
import logging

logger = logging.getLogger(__name__)


class InsightlyBackend(FieldMapperMixin, CRMBackend):
    api_version = 'v2.2'
    field_mapping = {
        'external_id': 'lead_id',
        'first_name': 'first_name',
        'last_name': 'last_name',
        'email': 'email_address',
        'phone': 'phone_number',
        'mobile_phone': 'mobile_phone_number',
        'street_address': 'address_street',
        'city': 'address_city',
        'state': 'address_state',
        'country': 'address_country',
    }

    def __init__(self, *args, **kwargs):
        super(InsightlyBackend, self).__init__(**kwargs)
        self._connection = None
        self._lead_statuses = None

    @property
    def connection(self):
        if self._connection:
            return self._connection

        self._connection = InsightlyAPI(api_key=self._auth_params['API_KEY'], version=self.api_version)
        return self._connection

    @property
    def lead_statuses(self):
        if self._lead_statuses:
            return self._lead_statuses

        self._lead_statuses = self.connection.read('leadstatuses', params={'include_converted': True})
        return self._lead_statuses

    def sync_lead(self, lead):
        crm_fields = self.apply_field_mapping(lead)

        if lead.external_status:
            # we have to extract the status ID of the status set on the lead
            try:
                crm_fields['lead_status_id'] = [l for l in self.lead_statuses if l['LEAD_STATUS'] == lead.external_status][0]['LEAD_STATUS_ID']
            except IndexError:
                raise ValueError('status {} isnt a valid lead status, correction required'.format(lead.external_status))

        if not lead.external_id:
            crm_fields.pop('lead_id')
            lead = self.connection.create('leads', crm_fields)
        else:
            lead = self.connection.update('leads', crm_fields)

        return lead['LEAD_ID']

    def delete_lead(self, lead_id):
        success = self.connection.delete('leads', id=lead_id)
        return success


# Generated by Django 2.0.2 on 2018-02-25 05:06

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('sales', '0001_initial'),
        ('crmify', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='lead',
            name='anchor',
            field=models.OneToOneField(default=1, on_delete=django.db.models.deletion.CASCADE, related_name='crm_lead', to='sales.UserProfile'),
            preserve_default=False,
        ),
    ]



from django.contrib import admin
from crmify import models


class LeadAdmin(admin.ModelAdmin):
    list_display = ('first_name', 'last_name', 'email', 'external_id', 'external_status', 'anchor', 'last_synced',)
    readonly_fields = ('external_id', 'last_synced',)


admin.site.register(models.Lead, LeadAdmin)


from json.decoder import JSONDecodeError
import os
import requests
import base64


class Insightly(object):
    base_url = 'https://api.insight.ly/'

    def __init__(self, version='v2.2', api_key=None):
        self.version = version
        self.api_key = api_key

    def get_url(self, path):
        return self.base_url + self.version + path

    def make_request(self, path, method='GET', **args):
        headers = {
            'Content-Type': 'application/json',
            'Authorization': 'Basic ' + str(base64.b64encode(self.api_key.encode('ascii')), encoding='ascii')
        }

        resp = requests.request(method, self.get_url(path), headers=headers, **args)

        if not resp.ok:
            resp.raise_for_status()

        try:
            return resp.json()
        except JSONDecodeError:
            return resp.content

    def read(self, resource, **args):
        return self.make_request('/' + resource, **args)

    def update(self, resource, data):
        return self.make_request('/' + resource, method='PUT', json=data)

    def create(self, resource, data):
        return self.make_request('/' + resource, method='POST', json=data)

    def delete(self, resource, id):
        return self.make_request('/' + os.path.join(resource, id), method='DELETE')


from django.db import models
from django.utils.timezone import now
from crmify.settings import crmify_settings
from crmify.backends import crm_backend
import logging

logger = logging.getLogger(__name__)


class CRMObject(models.Model):
    class Meta:
        abstract = True


class Lead(CRMObject):
    anchor = models.OneToOneField(crmify_settings.LEAD_MODEL, related_name='crm_lead', on_delete=models.CASCADE)

    lead_source = models.CharField(max_length=100, null=True, blank=True, default=None)
    first_name = models.CharField(max_length=150)
    last_name = models.CharField(max_length=150)
    email = models.CharField(max_length=200, null=True, blank=True, default=None)
    secondary_email = models.CharField(max_length=200, null=True, blank=True, default=None)
    phone = models.CharField(max_length=50, null=True, blank=True, default=None)
    mobile_phone = models.CharField(max_length=50, null=True, blank=True, default=None)
    dob = models.DateField(max_length=50, null=True, blank=True, default=None)
    street_address = models.CharField(max_length=200, null=True, blank=True, default=None)
    city = models.CharField(max_length=200, null=True, blank=True, default=None)
    state = models.CharField(max_length=50, null=True, blank=True, default=None)
    country = models.CharField(max_length=50, null=True, blank=True, default=None)

    external_id = models.CharField(max_length=200, unique=True)
    external_status = models.CharField(max_length=200, null=True, blank=True, default=None)
    last_synced = models.DateTimeField(null=True, blank=True, default=None)

    def sync_to_crm(self):
        logger.debug('now syncing lead to CRM')

        self.external_id = crm_backend.sync_lead(self)
        self.last_synced = now()

        return self.external_id

    def delete_from_crm(self):
        logger.debug('now deleting lead from CRM')
        crm_backend.delete_lead(self.external_id)

    def __str__(self):
        return """{} -- {} {} // {} ({})""".format(
            self.email, self.first_name, self.last_name, self.external_id, self.id
        )



# Generated by Django 2.0.2 on 2018-02-25 20:16

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('crmify', '0002_lead_anchor'),
    ]

    operations = [
        migrations.AddField(
            model_name='lead',
            name='last_synced',
            field=models.DateTimeField(blank=True, default=None, null=True),
        ),
    ]



from django.db.models.signals import post_save, post_delete, pre_save
from django.core.exceptions import ObjectDoesNotExist
from django.dispatch import receiver
from crmify.settings import crmify_settings
from crmify.models import Lead
import logging
import pdb

logger = logging.getLogger(__name__)


field_mapper = crmify_settings.LEAD_MODEL_FIELDMAPPER()


def get_possible_senders():
    """ the possible senders include the LEAD_MODEL itself, along with all many-to-one and one-to-one relations to the
    LEAD_MODEL (recursively)
    :return: `list` of `cls`
    """
    def rec_helper(cls, seen=None):
        seen = seen or []
        related_objects = [f for f in cls._meta.get_fields() if f.is_relation]

        all_possibles = [cls]
        frame_possibles = [o for o in related_objects if (o.many_to_one or o.one_to_one) and o.related_model is not Lead and o.related_model not in seen]
        for p in frame_possibles:
            all_possibles = all_possibles + rec_helper(p.related_model, seen=all_possibles)

        return list(set(all_possibles))

    return rec_helper(crmify_settings.LEAD_MODEL)

crm_senders = get_possible_senders()


def get_lead_model_instance_from_sender_instance(sender_instance):
    """ given a sender model instance, traverse the object graph to find the LEAD_MODEL instance associated with it.

    For instance, imagine we have objects like UserProfile -> User and then imagine that the UserProfile is the LEAD_MODEL
    and a User instance is saved. Therefore, the sender_instance will be a User, and the goal is to find the UserProfile(s)
    it is associated with to trigger a save on each

    :param sender:
    :return: `object` settings.LEAD_MODEL instance associated with the sender
    """
    for field in sender_instance._meta.get_fields():
        #TODO KNOWN LIMITATION: this will only work one layer deep.
        if field.is_relation and field.related_model is crmify_settings.LEAD_MODEL:
            try:
                return getattr(sender_instance, field.name)
            except ObjectDoesNotExist:
                # consider the example of UserProfile -> User. Say we modify a User's field before a UserProfile is created,
                # that would raise an ObjectDoesNotExist error
                return None


## Lead Syncing ##

@receiver(post_save)
def update_crmify_lead(sender=None, instance=None, created=None, **kwargs):
    if sender in crm_senders and sender is not crmify_settings.LEAD_MODEL:
        logger.debug('triggering save on lead model for related instance {}'.format(instance))
        lead_model_instance = get_lead_model_instance_from_sender_instance(instance)

        if lead_model_instance:
            lead_model_instance.save()
    elif sender == crmify_settings.LEAD_MODEL:
        if created:
            logger.debug('creating new lead for instance {}'.format(instance))
            lead = field_mapper.create_lead(instance)
        else:
            logger.debug('updating lead for instance {}'.format(instance))
            lead = Lead.objects.get(anchor=instance)
            lead = field_mapper.update_lead(lead, instance)

        logger.debug('lead is {}'.format(lead))


@receiver(post_delete, sender=crmify_settings.LEAD_MODEL)
def delete_crmify_lead(instance=None, **kwargs):
    logger.debug('deleting leads for instance {}'.format(instance))
    Lead.objects.filter(anchor=instance).delete()


@receiver(pre_save, sender=Lead)
def sync_crm_lead(instance=None, created=None, **kwargs):
    instance.sync_to_crm()


@receiver(post_delete, sender=Lead)
def delete_crm_lead(instance=None, **kwargs):
    instance.delete_from_crm()


if crmify_settings.LEAD_STATUS_MODEL:
    @receiver(post_save, sender=crmify_settings.LEAD_STATUS_MODEL)
    def sync_crm_lead_status(instance=None, created=None, **kwargs):
        lead_model, status = instance.lead_status()

        try:
            lead = lead_model.crm_lead
        except ObjectDoesNotExist:
            logger.warning('attempting to sync a crm lead status without a crm lead, abort')
            return

        lead.external_status = status
        lead.sync_to_crm()
        lead.save()


from crmify.mixins import FieldMapperMixin


class LeadModelFieldMapper(FieldMapperMixin):
    """ a LeadModelFieldMapper is a FieldMapper which specifically knows how to transform the fields of settings.LEAD_MODEL
    into a Lead object
    """
    def create_lead(self, instance):
        """ create a `Lead` object from the given object instance (an instance of the class given by settings.LEAD_MODEL)
        :param instance: `object` of class settings.LEAD_MODEL
        :return: `Lead` instance, as created from the given instance
        """
        from crmify.models import Lead

        lead_fields = self.apply_field_mapping(instance)
        lead_fields['anchor'] = instance
        return Lead.objects.create(**lead_fields)

    def update_lead(self, lead, instance):
        """ updates a `Lead` object from the given object instance
        :param lead: `Lead` instance to update
        :param instance: `object` of class settings.LEAD_MODEL
        :return: `Lead` instance as updated by the given instance
        """
        lead_fields = self.apply_field_mapping(instance)

        for f, val in lead_fields.items():
            setattr(lead, f, val)

        lead.save()

        return lead


class DjangoUserFieldMapper(LeadModelFieldMapper):
    field_mapping = {
        'first_name': 'first_name',
        'last_name': 'last_name',
        'email': 'email',
    }
    fallbacks = {
        'username': 'email'
    }




# Generated by Django 2.0.2 on 2018-02-25 02:37

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Lead',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('lead_source', models.CharField(blank=True, default=None, max_length=100, null=True)),
                ('first_name', models.CharField(max_length=150)),
                ('last_name', models.CharField(max_length=150)),
                ('email', models.CharField(blank=True, default=None, max_length=200, null=True)),
                ('secondary_email', models.CharField(blank=True, default=None, max_length=200, null=True)),
                ('phone', models.CharField(blank=True, default=None, max_length=50, null=True)),
                ('mobile_phone', models.CharField(blank=True, default=None, max_length=50, null=True)),
                ('dob', models.DateField(blank=True, default=None, max_length=50, null=True)),
                ('street_address', models.CharField(blank=True, default=None, max_length=200, null=True)),
                ('city', models.CharField(blank=True, default=None, max_length=200, null=True)),
                ('state', models.CharField(blank=True, default=None, max_length=50, null=True)),
                ('country', models.CharField(blank=True, default=None, max_length=50, null=True)),
                ('external_id', models.CharField(max_length=200, unique=True)),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



from crmify.models import Lead
from factory import faker
import factory


class LeadFactory(factory.DjangoModelFactory):
    email = faker.Faker('email')
    first_name = faker.Faker('name')
    last_name = faker.Faker('name')

    class Meta:
        model = Lead


from django.test import TestCase
from unittest import mock
from django.contrib.auth.models import User
from crmify.mappers import DjangoUserFieldMapper
from crmify import factories
from crmify.models import Lead


class FieldMapperTestCase(TestCase):
    def setUp(self):
        self.mapper = DjangoUserFieldMapper()
        self.instance = User(username='test@test.com', first_name='Bob', last_name='Testy', email='test@test.com')


class TestFieldMapperLookupValue(FieldMapperTestCase):
    def test_lookup_value_simple(self):
        instance = User(username='test@test.com', first_name='Bob', last_name='Testy')
        self.assertEqual(instance.username, self.mapper.lookup_value('username', instance))

    def test_lookup_value_nested(self):
        class NestedInner(object):
            def __init__(self, value):
                self.value = value

        class NestedOuter(object):
            def __init__(self):
                self.inner = NestedInner('test')

        instance = NestedOuter()
        self.assertEqual(instance.inner.value, self.mapper.lookup_value('inner__value', instance))


class TestFieldMapperFieldMapping(FieldMapperTestCase):
    def test_apply_field_mapping_to_simple_fields(self):
        self.assertDictEqual(
            self.mapper.apply_field_mapping(self.instance),
            {'first_name': self.instance.first_name, 'last_name': self.instance.last_name, 'email': self.instance.email}
        )

    def test_falls_back_to_username_for_email_when_no_email_set(self):
        self.instance.email = None
        self.assertDictEqual(
            self.mapper.apply_field_mapping(self.instance),
            {'first_name': self.instance.first_name, 'last_name': self.instance.last_name, 'email': self.instance.username}
        )

    def test_apply_field_mapping_to_fields_not_existing(self):
        class OtherClass(object):
            def __init__(self, foo=None):
                self.foo = foo

        instance = OtherClass(foo='bar')
        self.assertRaises(AttributeError, self.mapper.apply_field_mapping, instance)


class TestFieldMapperCreateLead(FieldMapperTestCase):
    def test_uses_result_from_apply_field_mapping_to_create_lead(self):
        self.mapper.apply_field_mapping = mock.MagicMock(return_value={})

        self.mapper.create_lead(self.instance)

        self.assertIsNone(self.mapper.apply_field_mapping.assert_called_once_with(self.instance))

    def test_returns_new_lead(self):
        new_lead = self.mapper.create_lead(self.instance)

        self.assertTrue(type(new_lead) is Lead)
        self.assertEqual(new_lead.email, self.instance.email)
        self.assertEqual(new_lead.first_name, self.instance.first_name)
        self.assertEqual(new_lead.last_name, self.instance.last_name)


class TestFieldMapperUpdateLead(FieldMapperTestCase):
    def setUp(self):
        super(TestFieldMapperUpdateLead, self).setUp()
        self.lead = factories.LeadFactory()

    def test_returns_updated_lead(self):
        updated_lead = self.mapper.update_lead(self.lead, self.instance)

        self.assertTrue(type(updated_lead) is Lead)
        self.assertEqual(updated_lead.email, self.instance.email)
        self.assertEqual(updated_lead.first_name, self.instance.first_name)
        self.assertEqual(updated_lead.last_name, self.instance.last_name)



from crmify.settings import crmify_settings


class FieldMapperMixin(object):
    field_mapping = {}
    fallbacks = {}

    def lookup_value(self, field, instance):
        """ nested value lookup for the given field on the given instance. For example, 'auth__username'
        :param field: `str` field to lookup
        :param instance: `object` to lookup field on
        :return: `value`
        """
        parts = field.split('__')
        field_name = parts[0]
        remainder = '__'.join(parts[1:])

        if not remainder:
            return getattr(instance, field_name)
        else:
            return self.lookup_value(remainder, getattr(instance, field_name))

    def apply_field_mapping(self, instance):
        fieldset = {}

        # model field, lead field
        for mfield, lfield in self.field_mapping.items():
            fieldset[lfield] = self.lookup_value(mfield, instance)

        for mfield, lfield in self.fallbacks.items():
            if fieldset.get(lfield) is None:
                fieldset[lfield] = self.lookup_value(mfield, instance)

        return fieldset


class LeadStatusMixin(object):
    """ subclasses of this Mixin should implement the single `lead_status` method. The result of this method will
    be used to signal lifecycle events in the Lead, which are then synced to the CRM
    """
    NEW_LEAD = crmify_settings.LEAD_NEW_STATUS
    CONVERTED_LEAD = crmify_settings.LEAD_CONVERTED_STATUS
    DEAD_LEAD = crmify_settings.LEAD_DEAD_STATUS

    def lead_status(self):
        raise NotImplementedError('to track lead statuses, you should implement the lead_status method')


class CRMBackend(object):
    def __init__(self, **auth_params):
        # for a variable like 'CRMIFY_BACKEND_AUTH_API_KEY', strip the CRMIFY_BACKEND_AUTH_ bit off
        self._auth_params = {k.replace('CRMIFY_BACKEND_AUTH_', ''): v for k, v in auth_params.items()}

    def sync_lead(self, lead):
        raise NotImplementedError

    def delete_lead(self, lead_id):
        raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import re
import shutil
import sys
from io import open

from setuptools import find_packages, setup

try:
    from pypandoc import convert_file

    def read_md(f):
        return convert_file(f, 'rst')
except ImportError:
    print("warning: pypandoc module not found, could not convert Markdown to RST")

    def read_md(f):
        return open(f, 'r', encoding='utf-8').read()


version = '0.3.0'


setup(
    name='django-crmify',
    version=version,
    url='https://github.com/MeanPug/django-crmify',
    license='Apache License 2.0',
    description='Connect your Django Application to your CRM of choice',
    long_description=read_md('README.md'),
    author='Bobby Steinbach',
    author_email='developers@meanpug.com',
    packages=find_packages(exclude=['tests*', 'examples*']),
    include_package_data=True,
    install_requires=[
        'requests>=2.18,<2.19'
    ],
    python_requires=">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*",
    zip_safe=False,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 1.10',
        'Framework :: Django :: 1.11',
        'Framework :: Django :: 2.0',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet :: WWW/HTTP',
    ]
)



""" Ripped from the django-rest-framework settings.py

Settings for CRMIFY library are all namespaced in the CRMIFY setting.
For example your project's `settings.py` file might look like this:

CRMIFY = {
    'BACKEND': 'crmify.backends.insightly.InsightlyBackend',
    'BACKEND_OPTIONS': {
        'LEAD_NEW_STATUS': '',
        'LEAD_DEAD_STATUS': ''
    }
}

This module provides the `crmify_settings` object, that is used to access
CRMIFY framework settings, checking for user settings first, then falling
back to the defaults.
"""
from __future__ import unicode_literals

from importlib import import_module

from django.conf import settings
from django.test.signals import setting_changed
from django.utils import six
import os


auth_keys = ['API_KEY', 'USERNAME', 'PASSWORD']

DEFAULTS = {
    'BACKEND': 'crmify.backends.insightly.InsightlyBackend',
    'BACKEND_AUTH': {k: os.environ.get('CRMIFY_BACKEND_AUTH_{}'.format(k)) for k in auth_keys},
    'LEAD_MODEL': 'django.contrib.auth.models.User',
    'LEAD_MODEL_FIELDMAPPER': 'crmify.mappers.DjangoUserFieldMapper',
    'LEAD_STATUS_MODEL': None,
    'LEAD_NEW_STATUS': 'NotContacted',
    'LEAD_CONVERTED_STATUS': 'Converted',
    'LEAD_DEAD_STATUS': 'Disqualified'
}


# List of settings that may be in string import notation.
IMPORT_STRINGS = (
    'BACKEND',
    'LEAD_MODEL_FIELDMAPPER',
    'LEAD_MODEL',
    'LEAD_STATUS_MODEL',
)


def perform_import(val, setting_name):
    """
    If the given setting is a string import notation,
    then perform the necessary import or imports.
    """
    if val is None:
        return None
    elif isinstance(val, six.string_types):
        return import_from_string(val, setting_name)
    elif isinstance(val, (list, tuple)):
        return [import_from_string(item, setting_name) for item in val]
    return val


def import_from_string(val, setting_name):
    """
    Attempt to import a class from a string representation.
    """
    try:
        # Nod to tastypie's use of importlib.
        module_path, class_name = val.rsplit('.', 1)
        module = import_module(module_path)
        return getattr(module, class_name)
    except (ImportError, AttributeError) as e:
        msg = "Could not import '%s' for API setting '%s'. %s: %s." % (val, setting_name, e.__class__.__name__, e)
        raise ImportError(msg)


class APISettings(object):
    """
    A settings object, that allows API settings to be accessed as properties.
    For example:

        from crmify.settings import crmify_settings
        print(crmify_settings.BACKEND)

    Any setting with string import paths will be automatically resolved
    and return the class, rather than the string literal.
    """
    def __init__(self, user_settings=None, defaults=None, import_strings=None):
        if user_settings:
            self._user_settings = self.__check_user_settings(user_settings)
        self.defaults = defaults or DEFAULTS
        self.import_strings = import_strings or IMPORT_STRINGS
        self._cached_attrs = set()

    @property
    def user_settings(self):
        if not hasattr(self, '_user_settings'):
            self._user_settings = getattr(settings, 'CRMIFY', {})
        return self._user_settings

    def __getattr__(self, attr):
        if attr not in self.defaults:
            raise AttributeError("Invalid API setting: '%s'" % attr)

        try:
            # Check if present in user settings
            val = self.user_settings[attr]
        except KeyError:
            # Fall back to defaults
            val = self.defaults[attr]

        # Coerce import strings into classes
        if attr in self.import_strings:
            val = perform_import(val, attr)

        # Cache the result
        self._cached_attrs.add(attr)
        setattr(self, attr, val)
        return val

    def __check_user_settings(self, user_settings):
        return user_settings

    def reload(self):
        for attr in self._cached_attrs:
            delattr(self, attr)
        self._cached_attrs.clear()
        if hasattr(self, '_user_settings'):
            delattr(self, '_user_settings')


crmify_settings = APISettings(None, DEFAULTS, IMPORT_STRINGS)


def reload_crmify_settings(*args, **kwargs):
    setting = kwargs['setting']

    if setting == 'CRMIFY':
        crmify_settings.reload()


setting_changed.connect(reload_crmify_settings)